/*
 This file is part of the digitalSTROM ds485-client library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-client library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM ds485-client library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM ds485-client library. If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>

#include <digitalSTROM/ds485-socket-interface.h>
#include "ds485-client-interface.h"
#include "ds485-client.h"

#ifdef GPL_FEATURE
#include <digitalSTROM/ds485-core.h>
#endif

/**
 * Opens a connection using the specified connection type
 *
 * @CONNSPEC:
 *
 *   serial:///dev/ttyUSB0 (GPL version only)
 *   tcp://localhost:8442
 *   udp://10.0.230.87:8442
 *   unix:///tmp/ds485.socket
 *
 * @RETURN:
 *   negative value on failure
 *   positive value is the client index
 */
ds485ClientHandle_t ds485_client_open(const char *connspec, int flags)
{
    int rval;
    ds485_client_t *cl;

    cl = (ds485_client_t *) malloc(sizeof(ds485_client_t));
    if (cl == NULL)
        return NULL;

    memset(cl, 0, sizeof(ds485_client_t));
    cl->connspec = strdup(connspec);
    cl->busState = DS485_STATE_ISOLATED;
    cl->flags = flags;

    pthread_mutex_init(&cl->syncSend.mutex, NULL);
    pthread_cond_init (&cl->syncSend.condition, NULL);
    cl->syncSend.state = STATE_IDLE;

    pthread_mutex_init(&cl->busQuery.mutex, NULL);
    pthread_cond_init (&cl->busQuery.condition, NULL);
    cl->busQuery.state = STATE_IDLE;

    pthread_mutex_init(&cl->openQuery.mutex, NULL);
    pthread_cond_init (&cl->openQuery.condition, NULL);
    cl->openQuery.state = STATE_IDLE;

#ifdef GPL_FEATURE
    if (!strncmp(connspec, "serial://", 9))
    {
        rval = ds485_serial_client_init(cl);
    }
    else
#endif
    if (!strncmp(connspec, "tcp://", 6))
    {
        rval = ds485_socket_client_init(cl);
    }
    else if (!strncmp(connspec, "udp://", 6))
    {
        rval = ds485_socket_client_init(cl);
    }
    else if (!strncmp(connspec, "unix://", 7))
    {
        rval = ds485_socket_client_init(cl);
    }
    else {
        rval = -1;
    }

    if (rval < 0) {
        free(cl->connspec);
        free(cl);
        return NULL;
    }

    dsid_t dsid;
    dsid.id[0] = 0x35;
    dsid.id[1] = 0x04;
    dsid.id[2] = 0x17;
    dsid.id[3] = 0x5f;
    dsid.id[4] = 0xe0;
    dsid.id[5] = 0;
    dsid.id[6] = 0;
    dsid.id[7] = 0xf0;
    dsid.id[8] = 0;
    dsid.id[9] = 0;
    dsid.id[10] = 0;
    dsid.id[11] = 1;

    cl->active = 1;

    {
        struct timespec ts;
        struct timeval now;

        pthread_mutex_lock(&cl->openQuery.mutex);
        cl->openQuery.state = STATE_WAITING;
        rval = cl->connInterface.open(cl, &dsid);

        gettimeofday(&now, NULL);
        ts.tv_sec = now.tv_sec + 10;
        ts.tv_nsec = now.tv_usec * 1000;

        if (cl->openQuery.state != STATE_RECEIVED) {
            pthread_cond_timedwait(&cl->openQuery.condition, &cl->openQuery.mutex, &ts);
        }
    }
    if (rval < 0 || cl->openQuery.state != STATE_RECEIVED) {
        pthread_mutex_unlock(&cl->openQuery.mutex);
        cl->active = 0;
        pthread_join(cl->connThread, NULL);
        return NULL;
    }
    cl->openQuery.state = STATE_CONNECTED;
    pthread_mutex_unlock(&cl->openQuery.mutex);

    return cl;
}

void ds485_cleanup(ds485_client_t *cl)
{
    pthread_mutex_destroy(&cl->syncSend.mutex);
    pthread_cond_destroy(&cl->syncSend.condition);

    pthread_mutex_destroy(&cl->busQuery.mutex);
    pthread_cond_destroy(&cl->busQuery.condition);

    pthread_mutex_destroy(&cl->openQuery.mutex);
    pthread_cond_destroy(&cl->openQuery.condition);

    free(cl->connspec);
    free(cl);
}

/**
 * Closes the session, frees all connection related resources
 */
int ds485_client_close(ds485ClientHandle_t handle)
{
    int rval;
    ds485_client_t *cl = (ds485_client_t *)handle;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    rval = cl->connInterface.close(cl);

    return rval;
}

/**
 * Query the list of devices on the dS485 bus
 */
int ds485_client_query_devices(ds485ClientHandle_t handle, dsid_t *device_list, int device_list_size)
{
    int rval;
    struct timespec ts;
    struct timeval now;
    ds485_client_t *cl = (ds485_client_t *)handle;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    pthread_mutex_lock(&cl->busQuery.mutex);
    cl->busQuery.state = STATE_WAITING;
    cl->busQuery.devices = device_list;
    cl->busQuery.length = device_list_size;

    rval = cl->connInterface.get_bus_devices(cl, device_list, device_list_size);

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + 10;
    ts.tv_nsec = now.tv_usec * 1000;

    if (cl->busQuery.state != STATE_RECEIVED) {
        pthread_cond_timedwait(&cl->busQuery.condition, &cl->busQuery.mutex, &ts);
    }

    if (cl->busQuery.state != STATE_RECEIVED)
    {
        rval = -1;
    }
    else
    {
        rval = cl->busQuery.length;
    }

    cl->busQuery.state = STATE_IDLE;
    pthread_mutex_unlock(&cl->busQuery.mutex);

    return rval;
}

/**
 * Query the current bus state
 */
int ds485_client_query_state(ds485ClientHandle_t handle, bus_state_t *state)
{
    int rval = -1;
    struct timespec ts;
    struct timeval now;
    ds485_client_t *cl = (ds485_client_t *)handle;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    pthread_mutex_lock(&cl->busQuery.mutex);
    cl->busQuery.state = STATE_WAITING;

    rval = cl->connInterface.get_bus_state(cl, state);

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + 2;
    ts.tv_nsec = now.tv_usec * 1000;

    if (cl->busQuery.state != STATE_RECEIVED) {
        pthread_cond_timedwait(&cl->busQuery.condition, &cl->busQuery.mutex, &ts);
    }

    if (cl->busQuery.state != STATE_RECEIVED)
    {
        rval = -1;
        *state = 0;
    }
    else
    {
        rval = 0;
        *state = cl->busQuery.length;
    }

    cl->busQuery.state = STATE_IDLE;
    pthread_mutex_unlock(&cl->busQuery.mutex);

    return rval;
}

/**
 * Send dM-API command
 */
int ds485_client_send_command(ds485ClientHandle_t handle, ds485_container_t *request)
{
    int rval;
    ds485_client_t *cl = (ds485_client_t *)handle;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    rval = cl->connInterface.send(cl, request);

    return rval;
}

/**
 * Send dM-API command, receive a single response packet
 */
int ds485_client_send_sync_command(ds485ClientHandle_t handle, ds485_container_t *request, ds485_container_t *response, int timeout)
{
    ds485_client_t *cl = (ds485_client_t *)handle;
    int rval;
    struct timespec ts;
    struct timeval now;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    pthread_mutex_lock(&cl->syncSend.mutex);
    response->sourceId = request->destinationId;
    response->containerType = DS485_CONTAINER_RESPONSE;
    response->data[0] = request->data[0];
    response->data[1] = request->data[1];
    response->transactionId = request->transactionId;
    cl->syncSend.response = *response;
    cl->syncSend.state = STATE_WAITING;

    rval = cl->connInterface.send(cl, request);
    if (rval < 0) {
        cl->syncSend.state = STATE_IDLE;
        pthread_mutex_unlock(&cl->syncSend.mutex);
        return rval;
    }

    if (IsBroadcastId(request->destinationId))
    {
        cl->syncSend.state = STATE_IDLE;
        pthread_mutex_unlock(&cl->syncSend.mutex);
        return NOTICE_CALL_TO_BROADCAST;
    }

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    if (cl->syncSend.state != STATE_RECEIVED) {
        pthread_cond_timedwait(&cl->syncSend.condition, &cl->syncSend.mutex, &ts);
    }

    if (cl->syncSend.state == STATE_RECEIVED)
    {
        *response = cl->syncSend.response;
        rval = 0;
    } else {
        rval = ERROR_RESPONSE_TIMEOUT;
    }

    cl->syncSend.state = STATE_IDLE;
    pthread_mutex_unlock(&cl->syncSend.mutex);

    return rval;
}

/**
 * Set and remove a reception handler for dS485 Packet's
 */
client_callback_container_t ds485_client_set_receive_callback(ds485ClientHandle_t handle, client_callback_container_t cb, void *data)
{
    ds485_client_t *cl = (ds485_client_t *)handle;
    if (cl == NULL) {
        return NULL;
    }
    client_callback_container_t prev_cb = cl->connInterface.responseCallback;
    cl->connInterface.responseCallback = cb;
    cl->connInterface.responseCallbackData = data;
    return prev_cb;
}

/**
 * Set and remove a callback handler to register for bus state changes
 */
client_callback_state_t ds485_client_set_state_callback(ds485ClientHandle_t handle, client_callback_state_t cb, void *data)
{
    ds485_client_t *cl = (ds485_client_t *)handle;
    if (cl == NULL) {
        return NULL;
    }
    client_callback_state_t prev_cb = cl->connInterface.stateCallback;
    cl->connInterface.stateCallback = cb;
    cl->connInterface.stateCallbackData = data;
    return prev_cb;
}

client_callback_bus_change_t ds485_client_set_bus_change_callback(ds485ClientHandle_t handle, client_callback_bus_change_t cb, void *data)
{
    ds485_client_t *cl = (ds485_client_t *)handle;
    if (cl == NULL) {
        return NULL;
    }
    client_callback_bus_change_t prev_cb = cl->connInterface.busChangeCallback;
    cl->connInterface.busChangeCallback = cb;
    cl->connInterface.busChangeCallbackData = data;
    return prev_cb;
}

int ds485_client_get_dsid(ds485ClientHandle_t handle, dsid_t *dsid)
{
    int rval = -1;
    struct timespec ts;
    struct timeval now;
    ds485_client_t *cl = (ds485_client_t *)handle;
    dsid_t local_dsid;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    pthread_mutex_lock(&cl->busQuery.mutex);
    cl->busQuery.state = STATE_WAITING;
    cl->busQuery.devices = &local_dsid;

    rval = cl->connInterface.get_dsid(cl);

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + 2;
    ts.tv_nsec = now.tv_usec * 1000;

    if (cl->busQuery.state != STATE_RECEIVED) {
        pthread_cond_timedwait(&cl->busQuery.condition, &cl->busQuery.mutex, &ts);
    }

    if (cl->busQuery.state != STATE_RECEIVED)
    {
        rval = ERROR_RESPONSE_TIMEOUT;
        SetNullId(*dsid);
    }
    else
    {
        rval = 0;
        *dsid = cl->busQuery.devices[0];
    }

    pthread_mutex_unlock(&cl->busQuery.mutex);
    cl->busQuery.state = STATE_IDLE;

    return rval;
}
